import numpy as np
class TreeReg:
    """A class that store the topology information of a graph in order to remove irregularities"""
    def __init__(self, i, j):
        """
        Initialise a tree node with the root of the tree node coordinate and an empty children node list

        :param i: X axis coodinates of the current tree node
        :type i: int
        :param j: Y axis coodinates of the current tree node
        :type j: int
        """
        self.plot = (i, j)
        self.children = []

    def add_children(self, i, j):
        """
        Instantiate a children node and add it to the list of the children nodes

        :param i: X axis coodinates of the root of the children node
        :type i: int
        :param j: Y axis coodinates of the root of the children node
        :type j: int
        """
        next_tree = TreeReg(i, j)
        self.children.append(next_tree)
        return next_tree

    def recursive_reg(self,A, i, j, n, tree, plot):
        """
        A recursive function that iterate through the continuous graph and store the children of each node as well as its number of ancestor

        :param A: skeleton of the segmentation
        :type A: np.float
        :param i: X axis coodinates of the current node
        :type i: int
        :param j: Y axis coodinates of the current node
        :type j: int
        :param n: number of ancestor of the current node
        :type n: int
        :param tree: current tree node
        :type tree: PVBM.GraphRegularisation.GraphRegularisation.TreeReg
        :param plot: A numpy array where the continuous navigated graph is saved
        :type plot: np.array

        """
        up = (i - 1, j)
        down = (i + 1, j)
        left = (i, j - 1)
        right = (i, j + 1)

        up_left = (i - 1, j - 1)
        up_right = (i - 1, j + 1)
        down_left = (i + 1, j - 1)
        down_right = (i + 1, j + 1)
        points = [up, down, left, right, up_left, up_right, down_left, down_right]
        children = np.sum([A[point] for point in points if
                           (point[0] >= 0 and point[1] >= 0 and point[0] < A.shape[0] and point[1] < A.shape[1])])
        if children >= 1:
            plot[i, j] = 1

        for point in points:
            if point[0] >= 0 and point[0] < A.shape[0] and point[1] < A.shape[1] and point[1] >= 0:
                if A[point] == 1:
                    tree__ = tree.add_children(point[0], point[1])
                    A[point] = 0
                    self.recursive_reg(A, point[0], point[1], n + 1, tree__, plot)

    def print_reg(self,tree, plot):
        """
        A recursive function that correct graphs irregularities by removing children branch which contains less than 10 pixels

        :type tree: PVBM.GraphRegularisation.GraphRegularisation.TreeReg
        :param plot: A numpy array where the corrected graph is saved
        :type plot: np.array

        """
        if len(tree.children) == 0:
            return 1
        else:
            n = 1 + sum([self.print_reg(child, plot) for child in tree.children])
            if n >= 10:
                plot[tree.plot] = 1
            return n


from PIL import Image,ImageFilter #Import Pillow library to load the images
import numpy as np #Import numpy
from skimage.morphology import skeletonize,square,dilation #Import functions to compute morphological operations
import os
import pathlib
path_to_save_datasets = "../PVBM_datasets"
from PVBM.Datasets import PVBM_Datasets
import sys
sys.setrecursionlimit(100000)

# dataset_downloader = PVBM_Datasets()
# dataset_downloader.download_dataset("Crop_HRF", path_to_save_datasets)
# dataset_downloader.download_dataset("INSPIRE", path_to_save_datasets)
# dataset_downloader.download_dataset("UNAF", path_to_save_datasets)
# print("Images downloaded successfully")
segmentation_path = list(pathlib.Path(path_to_save_datasets).glob("*/artery/*"))[10]
segmentation = Image.open(segmentation_path) #Open the segmentation
image_path = str(segmentation_path).replace("artery","images").replace("veins", "images")
image = Image.open(image_path)
# Depending on the quality of the segmentation, you would need to regularize (smooth) it more or less
#before computing the skeleton for instance by uncomment the following command
#segmentation = segmentation.filter(ImageFilter.ModeFilter(size=3))

segmentation = np.array(segmentation)/255 #Convert the segmentation to a numpy array with value 0 and 1
skeleton = skeletonize(segmentation)*1 # Compute the skeleton of the segmentation

from PVBM.DiscSegmenter import DiscSegmenter
segmenter = DiscSegmenter()
optic_disc = segmenter.segment(str(segmentation_path).replace("artery","images").replace("veins", "images"))
center, radius, roi, zones_ABC = segmenter.post_processing(optic_disc, max_roi_size = 600)
from PVBM.CentralRetinalAnalysis import CREVBMs
creVBMs = CREVBMs()
zone_A_ = zones_ABC[:,:,1]/255
zone_B_ = zones_ABC[:,:,0]/255
zone_C_ = zones_ABC[:,:,2]/255
outsideB = (zone_C_ - zone_B_)
segmentation_roi = (segmentation * outsideB)
skeleton_roi = (skeleton * outsideB)
out = creVBMs.compute_central_retinal_equivalents(segmentation_roi.copy(), skeleton_roi.copy(),center[0],center[1], radius, artery = True, Toplot = False )
1



import numpy as np
from skimage.morphology import skeletonize
from scipy.signal import convolve2d
from PVBM.helpers.tortuosity import compute_tortuosity
from PVBM.helpers.perimeter import compute_perimeter_
from PVBM.helpers.branching_angle import compute_angles_dictionary
#from PVBM.helpers.far import far
from PVBM.GraphRegularisation.GraphRegularisation import TreeReg
class GeometricalVBMs:
    """A class that can perform geometrical biomarker computation for a fundus image.
    """

    def extract_subgraphs(self,graphs, x_c, y_c):
        """
        Extract B, the a graph where each of the disconnected subgraph is labeled differently and D which contains the euclidian distance graph between each
        point in the graph and the optic disc.

        :param graphs: Original blood vessel segmentation graph
        :type graphs: array
        :param x_c: x axis of the optic disc center
        :type x_c: int
        :param y_c: y axis of the optic disc center
        :type y_c: int

        :return: B,D
        :rtype: tuple
        """
        B = np.zeros_like(graphs, dtype=np.float32)
        D = np.zeros_like(graphs, dtype=np.float32)
        n = 1
        for i in range(graphs.shape[0]):
            for j in range(graphs.shape[1]):
                if B[i, j] == 0 and graphs[i, j] == 1:
                    self.recursive_subgraph(graphs, B, D, i, j, n, x_c, y_c)
                    n += 1
        return B, D

    def recursive_subgraph(self,A, B, D, i, j, n, x_c, y_c):
        """
        Recursively extract the value within B and D.

        :param A: Original blood vessel segmentation graph
        :type A: array
        :param B: A graph where each of the disconnected subgraph is labeled differentl, which is initialized by a zeros matrix and recursively built
        :type B: array
        :param D: Euclidian Distance graph between each point in A and the optic disc center (x_c,y_c), which is initialized by a zeros matrix and recursively built
        :type D: array
        :param i: Current x axis location within the graph
        :type i: int
        :param j: Current y axis location within the graph
        :type j: int
        :param n: Current number of point distance since the optic disc
        :type n: int
        :param x_c: x axis of the optic disc center
        :type x_c: int
        :param y_c: y axis of the optic disc center
        :type y_c: int

        :return: B,D
        :rtype: tuple
        """
        up = (i - 1, j)
        down = (i + 1, j)
        left = (i, j - 1)
        right = (i, j + 1)

        up_left = (i - 1, j - 1)
        up_right = (i - 1, j + 1)
        down_left = (i + 1, j - 1)
        down_right = (i + 1, j + 1)
        points = [up, down, left, right, up_left, up_right, down_left, down_right]
        for point in points:
            if point[0] >= 0 and point[0] < B.shape[0] and point[1] < B.shape[1] and point[1] >= 0:
                if A[point] == 1:
                    B[point] = n
                    A[point] = 0
                    D[point] = ((y_c - point[0]) ** 2 + (x_c - point[1]) ** 2) ** 0.5
                    self.recursive_subgraph(A, B, D, point[0], point[1], n, x_c, y_c)

    def recursive_topology(self,A, B, i, j, n, max_radius, x_c, y_c, endpoints, interpoints, i_or, j_or, dico, bacount, bapos,
                        dist=0):
        """
        Recursively compute and analyze the topology of a segmented image.

        This function traverses a segmented image to identify endpoints and intersection points of a structure, and
        calculates various distance metrics. The results are stored in a dictionary for further analysis.

        :param A: The array representing the binary segmentation of the structure.
        :type A: np.array
        :param B: An auxiliary array used for processing.
        :type B: np.array
        :param i: The current x-coordinate.
        :type i: int
        :param j: The current y-coordinate.
        :type j: int
        :param n: The current recursion depth.
        :type n: int
        :param max_radius: The maximum allowed radius for the traversal.
        :type max_radius: float
        :param x_c: The x-coordinate of the center of the structure.
        :type x_c: int
        :param y_c: The y-coordinate of the center of the structure.
        :type y_c: int
        :param endpoints: The array to mark endpoints of the structure.
        :type endpoints: np.array
        :param interpoints: The array to mark intersection points of the structure.
        :type interpoints: np.array
        :param i_or: The original x-coordinate of the starting point.
        :type i_or: int
        :param j_or: The original y-coordinate of the starting point.
        :type j_or: int
        :param dico: A dictionary to store computed distance metrics.
        :type dico: dict
        :param bacount: A counter to track the number of recursive steps.
        :type bacount: int
        :param bapos: The backup position for the traversal.
        :type bapos: tuple or None
        :param dist: The cumulative distance of the traversal.
        :type dist: float

        :return: None
        """
        up = (i - 1, j)
        down = (i + 1, j)
        left = (i, j - 1)
        right = (i, j + 1)

        up_left = (i - 1, j - 1)
        up_right = (i - 1, j + 1)
        down_left = (i + 1, j - 1)
        down_right = (i + 1, j + 1)
        points = [up, down, left, right, up_left, up_right, down_left, down_right]
        children = np.sum([A[point] for point in points])
        distances = [1, 1, 1, 1, 2 ** 0.5, 2 ** 0.5, 2 ** 0.5, 2 ** 0.5]

        if ((y_c - i) ** 2 + (
                x_c - j) ** 2) ** 0.5 > max_radius:  # and np.mean(curtree.diameter_list)*6 <= 80  and ultimatum == False:
            endpoints[i, j] = 1
            true_dist = ((i_or - i) ** 2 + (j_or - j) ** 2) ** 0.5
            dico[(i_or, j_or, i, j)] = (dist, true_dist, true_dist / dist, bapos)
            i_or, j_or = i, j
            dist = 0
            return

        elif children == 0 and n >= 10:
            endpoints[i, j] = 1
            true_dist = ((i_or - i) ** 2 + (j_or - j) ** 2) ** 0.5
            dico[(i_or, j_or, i, j)] = (dist, true_dist, true_dist / dist, bapos)
            i_or, j_or = i, j
            dist = 0
            return

        elif children > 1 and n >= 10:
            interpoints[i, j] = 1
            true_dist = ((i_or - i) ** 2 + (j_or - j) ** 2) ** 0.5
            dico[(i_or, j_or, i, j)] = (dist, true_dist, true_dist / dist, bapos)
            i_or, j_or = i, j
            dist = 0
            n = 0
            bacount = 0
            bapos = None
        for point, distance in zip(points, distances):
            if point[0] >= 0 and point[0] < B.shape[0] and point[1] < B.shape[1] and point[1] >= 0:
                if A[point] == 1:
                    A[point] = 0
                    if bacount == 30:
                        bapos = point
                        # print("update bapos",bapos)
                        # print(i_or,j_or)
                    self.recursive_topology(A, B, point[0], point[1], n + 1, max_radius, x_c, y_c, endpoints, interpoints, i_or,
                                    j_or, dico, bacount + 1, bapos, distance + dist)


    def compute_geomVBMs(self,blood_vessel, skeleton, xc,yc, radius):
        """
        Compute various geometrical vascular biomarkers (VBMs) for a given blood vessel graph.

        This function analyzes the blood vessel segmentation and skeleton to extract several biomarkers such as area,
        tortuosity index, median tortuosity, overall length, median branching angle, and counts of start, end, and
        intersection points. It also provides visualizations of specific points on the graph.

        :param blood_vessel: Blood vessel segmentation containing binary values within {0,1}.
        :type blood_vessel: np.array
        :param skeleton: Blood vessel segmentation skeleton containing binary values within {0,1}.
        :type skeleton: np.array
        :param xc: X-axis coordinate of the optic disc center.
        :type xc: int
        :param yc: Y-axis coordinate of the optic disc center.
        :type yc: int
        :param radius: Radius in pixels of the optic disc.
        :type radius: int

        :return: A tuple containing:
            - A list of biomarkers [area, tortuosity index, median tortuosity, overall length, median branching angle, number of start points, number of end points, number of intersection points].
            - A tuple of visualizations (endpoints, interpoints, startpoints, angles_dico, dico).
        :rtype: Tuple[list, tuple]
        """
        ####Compute the area
        area = np.sum(blood_vessel)


        ## Extract the distances graphs
        B, D = self.extract_subgraphs(graphs=skeleton.copy(), x_c=xc, y_c=yc)

        ## Extract the starting points list by navigating through the skeleton graph
        starting_points = np.zeros((skeleton.shape[0], skeleton.shape[1]), dtype=float)
        for i in set(list(B.reshape(-1))) - {0}:
            mask = B == i
            if mask.sum() >= 50:
                min_index = (D * mask + (1 - mask) * 1e10).argmin()
                min_coordinates = np.unravel_index(min_index, D.shape)
                # print(((min_coordinates[0] - yc)**2 + (min_coordinates[1] - xc)**2)**0.5)
                if ((min_coordinates[0] - yc) ** 2 + (min_coordinates[1] - xc) ** 2) ** 0.5 < 100 + 1 * radius:
                    starting_points[min_coordinates[0], min_coordinates[1]] = 1
        starting_point_list = np.argwhere(starting_points == 1)


        ### Cleaning the skeleton graph irregularities
        B = np.zeros((blood_vessel.shape[0], blood_vessel.shape[1]))
        tree_reg_list = []
        plot = np.zeros((blood_vessel.shape[0], blood_vessel.shape[1]))
        for idx_start in starting_point_list:
            tree = TreeReg(idx_start[0], idx_start[1])
            tree.recursive_reg(skeleton.copy(), idx_start[0], idx_start[1], 0, tree, plot)
            tree_reg_list.append(tree)
        plot_list = []
        for tree_reg in tree_reg_list:
            plot = np.zeros((blood_vessel.shape[0], blood_vessel.shape[1]))
            tree_reg.print_reg(tree_reg, plot)
            plot_list.append(plot.copy())
        skoustideB_reg = np.sum(np.array(plot_list), axis=0)

        #####Initialise the endpoints, interpoints and startpoints array that will be used later for visualization
        endpoints = np.zeros((blood_vessel.shape[0], blood_vessel.shape[1]))
        interpoints = np.zeros((blood_vessel.shape[0], blood_vessel.shape[1]))
        startpoints = np.zeros((blood_vessel.shape[0], blood_vessel.shape[1]))

        #####Initialising a dictionary that will be used to store the topology of th graph
        dico = {}

        ####Navigating through the graph to fill the end,inter and startpoints array and the topology dico
        for idx_start in starting_point_list:
            # print("new index")
            # print(t)
            i, j = idx_start
            startpoints[i, j] = 1
            self.recursive_topology(skoustideB_reg.copy(), B, idx_start[0], idx_start[1], 1, np.inf, xc, yc, endpoints,
                            interpoints, i, j, dico, 0, None)

        #### Extracting the tortuosity and length using the topology of the graph
        chord_list = np.array([val[1] for val in dico.values()])
        arc_list = np.array([val[0] for val in dico.values()])
        TI = arc_list.sum() / chord_list.sum()
        medTor = np.median(arc_list / chord_list)
        ovlen = np.sum(arc_list)

        ####Filtering the potential double angles for the branching angles computations
        angles_dico = {}
        s = dico.keys()
        v = list(dico.values())
        for element, val in zip(s, v):
            angles_dico[(element[0], element[1])] = angles_dico.get((element[0], element[1]), []) + [val[3]]

        ####Storing all the branching angles value
        angles = []
        for key, value in angles_dico.items():
            b = key
            if len(value) > 1:
                a, c = value[0], value[1]
                if b != None and c != None and a != None:
                    b = np.array(b)
                    a = np.array(a)
                    c = np.array(c)
                    ba = a - b
                    bc = c - b

                    cosine_angle = np.dot(ba, bc) / (np.linalg.norm(ba) * np.linalg.norm(bc))
                    angle = np.arccos(cosine_angle)
                    angles.append(np.degrees(angle))
        ####Computing the median branching angles, the number of start/inter/endpoints
        medianba = np.median(angles)
        startp = len(starting_point_list)
        endp = endpoints.sum()
        interp = interpoints.sum()

        #### Return the biomarkers as well as the particular points visualisations
        return [area, TI, medTor, ovlen, medianba, startp, endp, interp], (endpoints,interpoints,startpoints, angles_dico, dico)






import gdown
import os
import zipfile


class PVBM_Datasets:
    """A class that downloads the PVBM datasets."""
    def __init__(self):
        self.file_ids = {
            "UNAF": "1IM5qUEARNp2RFpzKmILbdgasjLuJEIcX",
            "INSPIRE": "18TcmkuN_eZgM2Ph5XiX8x7_ejtKhA3qb",
            "Crop_HRF": "1QcozuK5yDyXbBkHqkbM5bxEkTGzkPDl3"
        }

    def download_dataset_from_google_drive(self, file_id, save_path):
        url = f'https://drive.google.com/uc?id={file_id}'
        gdown.download(url, save_path, quiet=False)
        print(f"Dataset downloaded and saved to {save_path}")

    def unzip_file(self, zip_file_path, extract_to_dir):
        """
        Unzip a file to the specified directory and delete the zip file after extraction.

        :param zip_file_path: Path to the zip file.
        :type zip_file_path: String

        :param extract_to_dir: Directory where the contents will be extracted.
        """
        with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
            zip_ref.extractall(extract_to_dir)
        print(f"Files extracted to {extract_to_dir}")

        os.remove(zip_file_path)
        print(f"Deleted the zip file: {zip_file_path}")

    def download_dataset(self, name, save_folder_path):
        """
        Download the PVBM datasets

        :param name: Name of the dataset to download. Need to be within {"UNAF", "INSPIRE", "Crop_HRF"}
        :param save_folder_path: Path to the folder where to store the downloaded datasets
        """
        if name in list(self.file_ids.keys()):
            zip_save_path = f"{name}.zip"
            os.makedirs(save_folder_path, exist_ok=True)
            # Call the function to download the dataset
            self.download_dataset_from_google_drive(self.file_ids[name], zip_save_path)
            # Call the function to unzip the file and delete the zip file
            self.unzip_file(zip_save_path, save_folder_path)
        else:
            print("Name should be within the following:", list(self.file_ids.keys()))


import numpy as np
from skimage.morphology import skeletonize,square,dilation

class Tree:
    """A class that store the topology information of a graph relevant for the CRE measurements"""
    def __init__(self, origin, startpoint):
        self.startpoint = startpoint
        self.children = []
        self.parents = []
        self.diameter_list = []
        self.endpoint = None
        self.plot_list = []
        self.origin = origin

    def update(self, diameter, plot):
        """
        Add a diameter and a plot to the list of the stored values

        :param diameter: diameter of the current measurement
        :type diameter: float
        :param plot: plot of the current measurement
        :type plot: np.array
        """
        self.diameter_list.append(diameter)
        self.plot_list.append(plot)

    def finished(self, endpoint):
        """
        Store the endpoint of the blood vessel to the list of the stored values

        :param endpoint: endpoint of the current measurement in x,y coordinate
        :type endpoint: (int,int)
        """
        self.endpoint = endpoint

    def add_children(self, startpoint):
        """
        Add a children node to the topology of the current node

        :param startpoint: startpoint of the new child node in x,y coordinate
        :type startpoint: (int,int)

        :next_tree: the newly initialised child node
        :rtype: PVBM.GraphCentralRetinalEquivalent.Tree
        """
        next_tree = Tree(self.origin, startpoint)
        next_tree.parents.append(self)
        self.children.append(next_tree)
        self.endpoint = startpoint
        return next_tree

    def set_children(self, children):
        """
        Store a children node to the list of the stored values

        :param children: children node to store in the list of the children nodes
        :type children: PVBM.GraphCentralRetinalEquivalent.Tree
        """
        self.children = children

    def remove_children(self):
        """
        Empty the list of the stored children nodes
        """
        self.saved_children = self.children[:]
        self.children = []
        return self

    def print_correct(self, saved_tree, max_d=0, first_tree=0):
        """
        Traverse the tree to find the subtree with the largest average diameter.

        This function computes the average diameter for the current tree node and its children, then identifies and
        returns the subtree with the largest average diameter encountered during the traversal.

        :param saved_tree: The subtree with the largest average diameter found so far.
        :type saved_tree: PVBM.GraphCentralRetinalEquivalent.Tree
        :param max_d: The maximum average diameter encountered so far.
        :type max_d: float
        :param first_tree: The initial tree node being processed.
        :type first_tree: PVBM.GraphCentralRetinalEquivalent.Tree

        :return: A tuple containing:
            - saved_tree (PVBM.GraphCentralRetinalEquivalent.Tree): The subtree with the largest average diameter.
            - max_d (float): The largest average diameter value.
            - first_tree (PVBM.GraphCentralRetinalEquivalent.Tree): The first subtree that met the diameter criteria.
        :rtype: Tuple[PVBM.GraphCentralRetinalEquivalent.Tree, float, PVBM.GraphCentralRetinalEquivalent.Tree]
        """
        if len(self.diameter_list) > 0:
            d = np.mean(self.diameter_list)
            if d > max_d:
                saved_tree = self
                max_d = d
                if first_tree == 0:
                    first_tree = self
        if len(self.children) >= 1:
            return max(
                [(saved_tree, max_d, first_tree)] + [children.print_correct(saved_tree, max_d, first_tree) for children
                                                     in self.children], key=lambda item: item[1])
        else:
            return saved_tree, max_d, first_tree

    def plotable_show_tree(self, l, shape, Toplot = False):
        """
        Compute and optionally visualize the tree's diameter information, then store the results in a list.

        This function computes various statistics about the tree's diameters and, if requested, creates a visualization
        of the tree. The results are stored in a dictionary, which is then appended to the provided list.

        :param l: A list to store the output dictionary containing the tree's diameter statistics and optional visualization.
        :type l: list
        :param shape: The shape of the array used for visualization.
        :type shape: tuple
        :param Toplot: A flag to decide if the visualization should be created and stored (setting it to True uses more RAM).
        :type Toplot: bool

        :return: None
        """
        if len(self.diameter_list) > 0:
            if Toplot:
                p = np.zeros((shape[0], shape[1]))
                tmp = self.plot_list  # [2:-2]
                for i in range(len(tmp)):
                    # print(i)
                    if i % 1 == 0:
                        p += tmp[i]
                tmp2 = np.zeros((shape[0], shape[1], 4))
                tmp2[:, :, 3] = dilation(p, square(3))
                tmp2[:, :, 1] = dilation(p, square(3))
            dico_output = {}
            if Toplot:
                dico_output['plot'] = tmp2
            dico_output['start'] = self.startpoint
            dico_output['Mean diameter'] = np.mean(self.diameter_list)  # [2:-2])
            dico_output['Median diameter'] = np.median(self.diameter_list)  # [2:-2])

            dico_output['Number of Measurement'] = len(self.diameter_list)  # [2:-2])
            dico_output['end'] = self.endpoint
            l.append(dico_output)
            return

    def compute_perpendicular_line(self,A, i, j, pi6, pj6, plotted_tmp, gt):
        """
        Compute the length of a perpendicular line extending from a given point and plot the results if required.

        This function calculates the length of a perpendicular line from a given point (i, j) based on the direction of
        the previous point (pi6, pj6). It optionally updates a provided plot with the calculated line. The function checks
        for boundaries and symmetry to ensure correct line computation.

        :param A: The 2D array representing the area where the perpendicular line is computed.
        :type A: np.array
        :param i: The current x-coordinate.
        :type i: int
        :param j: The current y-coordinate.
        :type j: int
        :param pi6: The previous x-coordinate.
        :type pi6: int
        :param pj6: The previous y-coordinate.
        :type pj6: int
        :param plotted_tmp: A temporary plot holder to be updated with the computed line.
        :type plotted_tmp: np.array or None
        :param gt: The ground truth 2D array used to determine boundary conditions.
        :type gt: np.array

        :return: The total length of the computed perpendicular line or None if a symmetry error is detected.
        :rtype: int or None
        """
        aoustideB = gt

        # Calculate the difference between the current and previous points
        direction = np.array([i - pi6, j - pj6])

        # Normalize the direction vector
        direction = direction / np.linalg.norm(direction)

        # Calculate the normal vector to the direction vector
        normal = np.array([-direction[1], direction[0]])

        # Temporary plot holder (clone of current plot)
        if plotted_tmp is not None:
            plotted_tmp_clone = plotted_tmp.copy()

        # Variables to keep track of distances in the positive and negative directions
        distance_pos = 0
        distance_neg = 0

        cond = True
        while cond:
            # Calculate positions in the positive direction
            pos_i = int(i + (distance_pos + 1) * normal[0])
            pos_j = int(j + (distance_pos + 1) * normal[1])

            # Calculate positions in the negative direction
            neg_i = int(i - (distance_neg + 1) * normal[0])
            neg_j = int(j - (distance_neg + 1) * normal[1])

            if (0 <= pos_i < A.shape[0] and 0 <= pos_j < A.shape[1] and
                    0 <= neg_i < A.shape[0] and 0 <= neg_j < A.shape[1]):

                pos_hit_boundary = aoustideB[pos_i, pos_j] == 1
                neg_hit_boundary = aoustideB[neg_i, neg_j] == 1

                if pos_hit_boundary and neg_hit_boundary:
                    if plotted_tmp is not None:
                        plotted_tmp_clone[pos_i, pos_j] = 1
                        plotted_tmp_clone[neg_i, neg_j] = 1
                    distance_pos += 1
                    distance_neg += 1
                elif pos_hit_boundary:
                    if plotted_tmp is not None:
                        plotted_tmp_clone[pos_i, pos_j] = 1
                    distance_pos += 1
                elif neg_hit_boundary:
                    if plotted_tmp is not None:
                        plotted_tmp_clone[neg_i, neg_j] = 1
                    distance_neg += 1
                else:
                    cond = False
            else:
                cond = False

        # If one side has extended significantly more than the other (symmetric error check)
        if abs(distance_pos - distance_neg) > 1:
            return None

        # If there's no symmetry error, commit the changes to the plot
        if plotted_tmp is not None:
            np.copyto(plotted_tmp, plotted_tmp_clone)
        # plt.imshow(inp[0]/255)
        # plt.imshow(zones/3)
        # plt.imshow(segmentation + gtvs.reshape((1472,1472,1)) + plotted_tmp.reshape((1472,1472,1)))
        # plt.show()
        total_distance = distance_pos + distance_neg
        return total_distance

    def recursive_CRE(self,A, B, D, i, j, n, pi, pj, pi2, pj2, pi3, pj3, pi4, pj4, pi5, pj5, pi6, pj6, plotted_tmp, i_or,
                        j_or, curtree, gt, x_c, y_c, radius_zone_C, p=0, ultimatum=False):
        """
        Recursively track the topology required to computed CRAE or CRVE equivalent for a given blood vessel graph.

        This function traverses a segmented image to compute the diameter of blood vessels and updates the current tree
        structure with the computed diameters. The function also handles visualization and ensures that the traversal
        respects the boundaries of the given region.

        :param A: The array representing the binary blood vessel segmentation.
        :type A: np.array
        :param B: An auxiliary array used for processing.
        :type B: np.array
        :param D: Another auxiliary array used for processing.
        :type D: np.array
        :param i: The current x-coordinate.
        :type i: int
        :param j: The current y-coordinate.
        :type j: int
        :param n: The current recursion depth.
        :type n: int
        :param pi: Previous x-coordinate.
        :type pi: int
        :param pj: Previous y-coordinate.
        :type pj: int
        :param pi2: Second previous x-coordinate.
        :type pi2: int
        :param pj2: Second previous y-coordinate.
        :type pi2: int
        :param pj2: int
        :param pi3: Third previous x-coordinate.
        :type pi3: int
        :param pj3: Third previous y-coordinate.
        :type pi3: int
        :param pj3: int
        :param pi4: Fourth previous x-coordinate.
        :type pi4: int
        :param pj4: Fourth previous y-coordinate.
        :type pi4: int
        :param pj4: int
        :param pi5: Fifth previous x-coordinate.
        :type pi5: int
        :param pj5: Fifth previous y-coordinate.
        :type pi5: int
        :param pj5: int
        :param pi6: Sixth previous x-coordinate.
        :type pi6: int
        :param pj6: Sixth previous y-coordinate.
        :type pi6: int
        :param pj6: int
        :param plotted_tmp: A temporary plot holder to be updated with the computed line.
        :type plotted_tmp: np.array or None
        :param i_or: The original x-coordinate of the starting point.
        :type i_or: int
        :param j_or: The original y-coordinate of the starting point.
        :type j_or: int
        :param curtree: The current tree structure being updated with diameters.
        :type curtree: PVBM.GraphCentralRetinalEquivalent.Tree
        :param gt: The ground truth 2D array used to determine boundary conditions.
        :type gt: np.array
        :param x_c: The x-coordinate of the optic disc center.
        :type x_c: int
        :param y_c: The y-coordinate of the optic disc center.
        :type y_c: int
        :param radius_zone_C: The radius of the optic disc zone.
        :type radius_zone_C: int
        :param p: An auxiliary parameter used for processing.
        :type p: int
        :param ultimatum: A flag used to determine specific processing conditions.
        :type ultimatum: bool

        :return: None
        """
        up = (i - 1, j)
        down = (i + 1, j)
        left = (i, j - 1)
        right = (i, j + 1)

        up_left = (i - 1, j - 1)
        up_right = (i - 1, j + 1)
        down_left = (i + 1, j - 1)
        down_right = (i + 1, j + 1)
        points = [up, down, left, right, up_left, up_right, down_left, down_right]
        children = np.sum([A[point] for point in points])
        diameter = 0
        if n >= 10:
            if pi6 != None:
                if plotted_tmp is not None:
                    prev_plot = plotted_tmp.copy()
                diameter = self.compute_perpendicular_line(A, i, j, pi6, pj6, plotted_tmp, gt)
                if diameter is not None:
                    p_up = None
                    if plotted_tmp is not None:
                        p_up = plotted_tmp - prev_plot
                    curtree.update(diameter, p_up)
        else:
            curtree = curtree.add_children((i, j))

        if ((x_c - i) ** 2 + (
                y_c - j) ** 2) ** 0.5 > radius_zone_C:  # and np.mean(curtree.diameter_list)*6 <= 80  and ultimatum == False:
            curtree.finished((i, j))
            return

        # elif (children==0 or children >1) and len(curtree.diameter_list) > 30: #and ultimatum == True:
        #    curtree.finished((i,j))
        #    return

        elif children == 0:
            curtree.finished((i, j))
            return

        elif children > 1:
            if np.mean(curtree.diameter_list) * 6 > 80 and len(curtree.diameter_list) > 30:
                # ultimatum = True
                todo = 0
            curtree = curtree.add_children((i, j))
            if plotted_tmp is not None:
                plotted_tmp = np.zeros((A.shape[0], A.shape[1]))
            n = 0
            p += 1

        for point in points:
            if point[0] >= 0 and point[0] < B.shape[0] and point[1] < B.shape[1] and point[1] >= 0:
                if A[point] == 1:
                    A[point] = 0
                    self.recursive_CRE(A, B, D, point[0], point[1], n + 1, i, j, pi, pj, pi2, pj2, pi3, pj3, pi4, pj4, pi5,
                                    pj5, plotted_tmp, i_or, j_or, curtree, gt, x_c, y_c, radius_zone_C, p, ultimatum)